/* TCPechod.c - main, TCPechod
    cesar kallas
    02099224
    cesarkallas at gmx.net
*/

#define	_USE_BSD
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/errno.h>
#include <netinet/in.h>
#include <sys/signal.h>		/* for kill(), sigsuspend(), others */
#include <sys/ipc.h>		/* for shmget(),shmat(), and shmctl() */
#include <sys/shm.h>		/* for shmget(),shmat(), and shmctl()  */
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define	QLEN		  32	/* maximum connection queue length	*/
#define	BUFSIZE		4096

extern int	errno;

void	reaper(int);
int	TCPechod(int fd);
int	errexit(const char *format, ...);
int	passiveTCP(const char *service, int qlen);

/*
 * SHARED_MEMORY_ID, e uma chave arbitraria, correspondendo a um numero 
 * qualquer que deve ser unico. Se alguem estiver usando essa mesma chave, 
 * sera necessario mudar esses numero!
 */
#define SHARED_MEMORY_ID 1234

typedef struct {
	int numero_filhos;
	float tempo_total;
} time_stats_t;

time_stats_t *g_time_stats;

/*------------------------------------------------------------------------
 * main - Concurrent TCP server for ECHO service
 *------------------------------------------------------------------------
 */
int
main(int argc, char *argv[])
{
	char	*service = "echo";	/* service name or port number	*/
	struct	sockaddr_in fsin;	/* the address of a client	*/
	unsigned int	alen;		/* length of client's address	*/
	int	msock;			/* master server socket		*/
	int	ssock;			/* slave server socket		*/
	struct timeval t_inicial;	/* time */
	struct timeval t_final;		
	float t_diferenca;
	
	/* 
         * Variaveis relacionadas com memoria compartilhada
         */
    int shm_id;
    key_t key = SHARED_MEMORY_ID;
    char *tmp_addr;
	
	switch (argc) {
	case	1:
		break;
	case	2:
		service = argv[1];
		break;
	default:
		errexit("usage: TCPechod [port]\n");
	}

	/*
	 * O estabelecimento da mascara de sinais
	 */
	sigset_t sigset;

    /*
         * Cria a memoria compartilhada
         */
	if( (shm_id = shmget(key, sizeof(time_stats_t), IPC_CREAT | 0666)) == -1 ) {
		fprintf(stderr,"Impossivel criar a memoria compartilhada!\n");
		exit(1);
	}

	/*
	 * Associa ao segmento de memoria compartilhada
	 */
	if( (tmp_addr = (char *)shmat(shm_id, NULL, 0)) == (char *)-1 ) {
		fprintf(stderr,"Impossivel associar ao segmento de memoria compartilhada!\n");
		exit(1);
	}

	/*
	 * Estabelecimento do ponteiro g_time_stats
	 */
	g_time_stats = (time_stats_t *)tmp_addr;


	/*
	 * Estabelece uma mascara de sinais default que boqueia SIGUSR1,
         * que e o sinal que ira ser enviado entre os filhos. Ambos os
	 * filhos herdam essa mascara.
	 */
	sigemptyset( &sigset );
	sigaddset( &sigset, SIGUSR1 );
	sigprocmask( SIG_BLOCK, &sigset, NULL );
	g_time_stats->numero_filhos=0;
	g_time_stats->tempo_total=0;
	
	msock = passiveTCP(service, QLEN);

	(void) signal(SIGCHLD, reaper);

	while (1) {
		alen = sizeof(fsin);
		ssock = accept(msock, (struct sockaddr *)&fsin, &alen);
		gettimeofday(&t_inicial, NULL );
		if (ssock < 0) {
			if (errno == EINTR)
				continue;
			errexit("accept: %s\n", strerror(errno));
		}
		
		switch (fork()) {
		case 0: { 	/* child */
		    close(msock);
			TCPechod(ssock);
			gettimeofday(&t_final, NULL);
			g_time_stats->numero_filhos += 1;
			t_diferenca = (float)(t_final.tv_sec  - t_inicial.tv_sec);
			t_diferenca += (t_final.tv_usec - t_inicial.tv_usec)/(float)1000000;
			g_time_stats->tempo_total += t_diferenca;
			
			printf("Client %.5d: %s:%d  t: %f   tt: %f   tm: %f\n",
				g_time_stats->numero_filhos,
				inet_ntoa(fsin.sin_addr),
				ntohs(fsin.sin_port),
				t_diferenca,
				g_time_stats->tempo_total,
				g_time_stats->tempo_total / g_time_stats->numero_filhos);
			exit(0);
		}
		default:	/* parent */
			(void) close(ssock);
			/*
			* Remove a memoria compartilhda
			*/
            if( shmctl(shm_id,IPC_RMID,NULL) != 0 )
				fprintf(stderr,"Impossivel remover a memoria compartilhada!\n");
			break;
		case -1:
			errexit("fork: %s\n", strerror(errno));
		}
	}
}

/*------------------------------------------------------------------------
 * TCPechod - echo data until end of file
 *------------------------------------------------------------------------
 */
int
TCPechod(int fd)
{
	char	buf[BUFSIZ];
	int	cc;

	while (cc = read(fd, buf, sizeof buf)) {
		if (cc < 0)
			errexit("echo read: %s\n", strerror(errno));
		if (write(fd, buf, cc) < 0)
			errexit("echo write: %s\n", strerror(errno));
	}
	return 0;
}

/*------------------------------------------------------------------------
 * reaper - clean up zombie children
 *------------------------------------------------------------------------
 */
void
reaper(int sig)
{
	int	status;

	while (wait3(&status, WNOHANG, (struct rusage *)0) >= 0)
		/* empty */;
}
